@@ -138,6 +138,61 @@ def get_existing(googleads_client,customer_id):
 
         return df
 
+def get_existing_keywords(googleads_client,customer_id):
+    camps = {
+        'camp_name': list(),
+        'camp_id': list(),
+        'camp_status': list(),
+        'adgroup_name': list(),
+        'adgroup_status': list(),
+        'adgroup_id': list(),
+        'keyword_name': list(),
+        'keyword_type': list(),
+    }
+    ga_service = googleads_client.get_service("GoogleAdsService")
+    query = """
+        SELECT
+          ad_group.id,
+          ad_group.name,
+          ad_group.status,
+          campaign.id,
+          campaign.name,
+          campaign.status,
+          ad_group_criterion.type,
+          ad_group_criterion.status,
+          ad_group_criterion.criterion_id,
+          ad_group_criterion.keyword.text,
+          ad_group_criterion.keyword.match_type
+        FROM ad_group_criterion
+        WHERE ad_group_criterion.type = KEYWORD"""
+
+    search_request = googleads_client.get_type("SearchGoogleAdsRequest")
+    search_request.customer_id = customer_id
+    search_request.query = query
+    search_request.page_size = _DEFAULT_PAGE_SIZE
+
+    results = ga_service.search(request=search_request)
+    for row in results:
+        ad_group = row.ad_group
+        ad_group_criterion = row.ad_group_criterion
+        keyword = row.ad_group_criterion.keyword
+        print(
+            f'Keyword with text "{keyword.text}", match type '
+            f"{keyword.match_type}, criteria type "
+            f"{ad_group_criterion.type_}, and ID "
+            f"{ad_group_criterion.criterion_id} was found in ad group "
+            f"with ID {ad_group.id}.")
+        camps['camp_name'].append(row.campaign.name)
+        camps['camp_id'].append(row.campaign.id)
+        camps['camp_status'].append(row.campaign.status.name)
+        camps['adgroup_name'].append(row.ad_group.name)
+        camps['adgroup_id'].append(row.ad_group.id)
+        camps['adgroup_status'].append(row.ad_group.status.name)
+        camps['keyword_name'].append(keyword.text)
+        camps['keyword_type'].append(keyword.match_type.name)
+    df = pd.DataFrame(camps)
+    return df
+
 def create_keyword(googleads_client,customer_id,ad_group_id, keyword_text, kw_type):
     ad_group_service = googleads_client.get_service("AdGroupService")
     ad_group_criterion_service = googleads_client.get_service(
@@ -282,6 +337,180 @@ def create_campaign(googleads_client,customer_id,campaignName, budgetName, budge
     except GoogleAdsException as ex:
         _handle_googleads_exception(ex)
 
+def get_all_ads(client,customer_id):
+    query = """
+    SELECT
+    ad_group.id,
+    campaign.id,
+    ad_group_ad.ad.id,
+    ad_group_ad.ad.responsive_search_ad.headlines,
+    ad_group_ad.ad.responsive_search_ad.descriptions,
+    ad_group_ad.ad.final_urls,
+    ad_group_ad.ad.responsive_search_ad.path1,
+    ad_group_ad.ad.responsive_search_ad.path2,
+    ad_group_ad.status
+    FROM ad_group_ad
+    WHERE ad_group_ad.ad.type = RESPONSIVE_SEARCH_AD
+    """
+
+    df_dict = {
+        'adgroup_id': list(),
+        'campaign_id': list(),
+        'adgroup_ad_id': list(),
+        'headline_keywords': list(),
+        'ad_description': list(),
+        'final_url': list(),
+        'path1': list(),
+        'path2': list()
+    }
+
+    ga_service = client.get_service("GoogleAdsService")
+    search_request = client.get_type("SearchGoogleAdsStreamRequest")
+    search_request.customer_id = customer_id
+    search_request.query = query
+    stream = ga_service.search_stream(search_request)
+    for batch in stream:
+        for row in batch.results:
+            temp = list()
+            temp2 = list()
+            temp3 = list()
+            adgroup = row.ad_group
+            adgroup_ad = row.ad_group_ad
+            campaign = row.campaign
+
+            for headline in adgroup_ad.ad.responsive_search_ad.headlines:
+                temp.append(headline.text)
+            for descriptions in adgroup_ad.ad.responsive_search_ad.descriptions:
+                temp2.append(descriptions.text)
+            for final_url in adgroup_ad.ad.final_urls:
+                temp3.append(final_url)
+
+            df_dict['adgroup_id'].append(adgroup.id)
+            df_dict['campaign_id'].append(campaign.id)
+            df_dict['adgroup_ad_id'].append(adgroup_ad.ad.id)
+            df_dict['headline_keywords'].append(temp)
+            df_dict['ad_description'].append(temp2)
+            df_dict['final_url'].append(temp3)
+            df_dict['path1'].append(adgroup_ad.ad.responsive_search_ad.path1)
+            df_dict['path2'].append(adgroup_ad.ad.responsive_search_ad.path2)
+
+    df = pd.DataFrame.from_dict(df_dict)
+    return df
+
+def get_ads(client,customer_id,ad_group_id):
+    query = f"""
+    SELECT
+    ad_group.id,
+    campaign.id,
+    ad_group_ad.ad.id,
+    ad_group_ad.ad.responsive_search_ad.headlines,
+    ad_group_ad.ad.responsive_search_ad.descriptions,
+    ad_group_ad.ad.final_urls,
+    ad_group_ad.ad.responsive_search_ad.path1,
+    ad_group_ad.ad.responsive_search_ad.path2,
+    ad_group_ad.status
+    FROM ad_group_ad
+    WHERE ad_group_ad.ad.type = RESPONSIVE_SEARCH_AD AND ad_group.id == {ad_group_id}
+    """
+
+    df_dict = {
+        'adgroup_id': list(),
+        'campaign_id': list(),
+        'adgroup_ad_id': list(),
+        'headline_keywords': list(),
+        'ad_description': list(),
+        'final_url': list(),
+        'path1': list(),
+        'path2': list()
+    }
+
+    ga_service = client.get_service("GoogleAdsService")
+    search_request = client.get_type("SearchGoogleAdsStreamRequest")
+    search_request.customer_id = customer_id
+    search_request.query = query
+    stream = ga_service.search_stream(search_request)
+    for batch in stream:
+        for row in batch.results:
+            temp = list()
+            temp2 = list()
+            temp3 = list()
+            adgroup = row.ad_group
+            adgroup_ad = row.ad_group_ad
+            campaign = row.campaign
+
+            for headline in adgroup_ad.ad.responsive_search_ad.headlines:
+                temp.append(headline.text)
+            for descriptions in adgroup_ad.ad.responsive_search_ad.descriptions:
+                temp2.append(descriptions.text)
+            for final_url in adgroup_ad.ad.final_urls:
+                temp3.append(final_url)
+
+            df_dict['adgroup_id'].append(adgroup.id)
+            df_dict['campaign_id'].append(campaign.id)
+            df_dict['adgroup_ad_id'].append(adgroup_ad.ad.id)
+            df_dict['headline_keywords'].append(temp)
+            df_dict['ad_description'].append(temp2)
+            df_dict['final_url'].append(temp3)
+            df_dict['path1'].append(adgroup_ad.ad.responsive_search_ad.path1)
+            df_dict['path2'].append(adgroup_ad.ad.responsive_search_ad.path2)
+
+    df = pd.DataFrame.from_dict(df_dict)
+    return df
+
+def create_ad(client, customer_id, ad_group_id, final_url, headlines, descriptions, path1, path2):
+
+    def _create_ad_text_asset(client, text, pinned_field=None):
+        """Create an AdTextAsset."""
+        ad_text_asset = client.get_type("AdTextAsset")
+        ad_text_asset.text = text
+        if pinned_field:
+            ad_text_asset.pinned_field = pinned_field
+        return ad_text_asset
+    
+    
+    ad_group_ad_service = client.get_service("AdGroupAdService")
+    ad_group_service = client.get_service("AdGroupService")
+
+    # Create the ad group ad.
+    ad_group_ad_operation = client.get_type("AdGroupAdOperation")
+    ad_group_ad = ad_group_ad_operation.create
+    ad_group_ad.status = client.enums.AdGroupAdStatusEnum.PAUSED
+    ad_group_ad.ad_group = ad_group_service.ad_group_path(
+        customer_id, ad_group_id
+    )
+
+    # Set responsive search ad info.
+    ad_group_ad.ad.final_urls.append(final_url)
+
+    # Set a pinning to always choose this asset for HEADLINE_1. Pinning is
+    # optional; if no pinning is set, then headlines and descriptions will be
+    # rotated and the ones that perform best will be used more often.
+    # served_asset_enum = client.enums.ServedAssetFieldTypeEnum.HEADLINE_1
+    heads = [_create_ad_text_asset(client, headline) for headline in headlines]
+    print(heads)
+
+    ad_group_ad.ad.responsive_search_ad.headlines.extend(
+        heads
+    )
+    descs = [_create_ad_text_asset(client, headline) for headline in descriptions]
+    ad_group_ad.ad.responsive_search_ad.descriptions.extend(
+        descs
+    )
+    print(descs)
+    ad_group_ad.ad.responsive_search_ad.path1 = path1
+    ad_group_ad.ad.responsive_search_ad.path2 = path2
+
+    # Send a request to the server to add a responsive search ad.
+    ad_group_ad_response = ad_group_ad_service.mutate_ad_group_ads(
+        customer_id=customer_id, operations=[ad_group_ad_operation]
+    )
+
+    for result in ad_group_ad_response.results:
+        print(
+            f"Created responsive search ad with resource name "
+            f'"{result.resource_name}".'
+        )
+
 def remove_campaign(client, customer_id, campaign_id):
     campaign_service = client.get_service("CampaignService")
     campaign_operation = client.get_type("CampaignOperation")
