@@ -3,7 +3,7 @@ from google.ads.googleads.client import GoogleAdsClient
 from autoads.gads import remove_adgroup,remove_campaign
 
 # from autoads.gads remove_campaign
-path = 'google-ads.yaml'
+path = 'data/google-ads.yaml'
 customer_id = '8306215642' #google ads customer id
 
 # latest csv file
